@@ -1,5 +1,5 @@
 /* Sentence handling.
-   Copyright (C) 2015 Free Software Foundation, Inc.
+   Copyright (C) 2015-2024 Free Software Foundation, Inc.
    Written by Daiki Ueno <ueno@gnu.org>, 2015.
 
    This program is free software: you can redistribute it and/or modify
@@ -28,11 +28,12 @@ extern "C" {
    end of sentence.  */
 extern DLL_VARIABLE int sentence_end_required_spaces;
 
-/* Locate the position of a sentence end marker (a period, a question
-   mark, etc) in a null-terminated string STR.  If there is no
-   sentence end marker found in STR, return a pointer to the null byte
-   at the end of STR.  ENDING_CHARP is a return location of the end
-   marker character.  */
+/* Locates the position of a sentence end marker (a period, a question
+   mark, etc.) in a null-terminated string STR.
+   If found, it returns a pointer to this marker and sets *ENDING_CHARP
+   to this marker character.
+   If not found, it returns a pointer to the NUL byte at the end of STR
+   and sets *ENDING_CHARP to U+FFFD.  */
 extern const char *sentence_end (const char *string, ucs4_t *ending_charp);
 
 #ifdef __cplusplus
